#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include "types.h"
#include "netio.h"
#include "fileio.h"
#include "globals.h"
#include "logfun.h"

#define INTBUFSIZ	8192	/* read/write ahead buffer size */

static CHAR rahBuffer[INTBUFSIZ];


DOUBLE getCountDM (count_t *counter)
{
    return ((DOUBLE)(*counter))/1048576.0;    
}


INT getXferBuffer (FILE *fd, CHAR *buf, INT reqlen, INT mode, count_t *iocount)
{
    static CHAR *next;
    static INT avail = 0, lfdone = 0; 
    INT actlen = 0;

    while (reqlen > 0) {

        if (avail > 0) {	/* still someting buffered ? */
            INT siz;
            siz = (reqlen > avail) ? avail : reqlen;
            if (mode == RM_NETASCII) {	/* convert <lf> to <cr><lf> */
		while (siz--) {
                    if (*next == '\n' && !lfdone) {
                        *buf++  = '\r';
                        lfdone  = 1;
                    }
                    else {
                        *buf++  = *next++;
                        lfdone  = 0;
                        avail--;
                    }
                    actlen++;
                    reqlen--;
                }
            }
            else {		/* copy largest possible chunk */
                memcpy(buf, next, siz);
                actlen += siz;
                reqlen -= siz;
                avail  -= siz;
                next   += siz;
                buf    += siz;
            }
        }
        else {			/* get next chunk from disk */
            avail = fread (rahBuffer, 1, sizeof(rahBuffer), fd);
            if (avail <= 0) {
                if (feof (fd))
                    break;	/* end-of-file: return partial buffer */
                else if (errno == EINTR) {
                    if (isAborted)	/* singnal */
                        return -1;
                    else
                        avail = 0;
                }
                else
                    return -1;	/* read error: just quit */
            }
            next = rahBuffer;
        }
    }
    iocount += (count_t)actlen;
    return actlen;
}

INT putXferBuffer (FILE *fd, CHAR *buf, INT reqlen, INT mode, INT flush,
                   count_t *iocount)
{
    static CHAR *next;
    static INT avail = 0, ignorecr = 0, outlen = 0; 
    INT actlen = reqlen;

    while (reqlen > 0) {
        if (avail > 0) {	/* still someting buffered ? */

            INT siz = (reqlen > avail) ? avail : reqlen;

            if (mode == RM_NETASCII) {	/* convert <cr><lf> to <lf> */

		while (siz--) {
                    if (*buf == '\r') {
                        ignorecr = 1;
                        buf++;
                        reqlen--;
                        continue;
                    }
                    if (ignorecr) {
                        ignorecr = 0;
                        if (*buf != '\n') {
                            *next++  = '\r';
                            avail--;
                            outlen++;
                            break;
                        }
                    }
                    *next++  = *buf++;
                    avail--;
                    outlen++;
                    reqlen--;
                }
            }
            else {		/* copy largest possible chunk */

                memcpy(next, buf, siz);
                outlen += siz;
                reqlen -= siz;
                avail  -= siz;
                next   += siz;
                buf    += siz;
            }
        }
        else {			/* write next chunk to disk */
            if (outlen)
                if ((INT)fwrite(rahBuffer, 1, outlen, fd) != outlen)
                    return -1;		/* write error: just quit */

            next   = rahBuffer;
            avail  = sizeof(rahBuffer);
            outlen = 0;
        }
    }

    if (flush) {
        if (outlen)
            if ((INT)fwrite (rahBuffer, 1, outlen, fd) != outlen)
                return -1;	/* write error: just quit */

        next   = rahBuffer;
        avail  = sizeof(rahBuffer);
        outlen = 0;
    }

    iocount += (count_t)actlen;
    return actlen;
}


#ifdef TEST

#include "globals.h"
#include "logfun.h"
#define TESTFILE "io_test.dat"

static FILE *testfd;

VOID mkTestfile(VOID)
{
    remove (TESTFILE);
    if ((testfd = fopen(TESTFILE, "w+")) == NULL) {
        printf("Cannot open test file -- exiting\n");
        exit(1);
    }

}

VOID removeTestfile (VOID)
{
    fclose(testfd);
    remove(TESTFILE);
}

static VOID filbuf (char *buf, INT size, INT mark)
{
    INT i, j;

    if (size <= 4) {
        sprintf(buf,"%3d",mark % 100);
        buf[3] = '\n';
    }
    else {
        sprintf(buf,"%3d ",mark % 100);
        for (i = 4; i < 62; i++)
            buf[i] = '-';
        buf[62] = '\r';
        buf[63] = '\n';

        if (size > 64) {
            for (i = 64; i < size; i++)
            if ((i % 64) == 63)
                buf[i] = '\n';
            else if ((i % 64) == 62)
                buf[i] = '\r';
            else if ((i % 64) < 3)
                buf[i] = ' ';
            else {
                j = (i % 16);
                buf[i] = (j > 9) ? (j - 10 + 'a') : j + '0';
            }
        }
    }
}


static INT docompare (INT seg, char *buf1, char *buf2, INT len)
{
    INT i, s, e ;

    for (i = 0; i < len; i++) {
        if(buf1[i] != buf2[i]) {
            printf ("Data mismatch in segment %d (%d) offset %d: file:[0x%2.2x] comp:[0x%2.2x]\n",
                    seg, len, i, (int)buf1[i], (int)buf2[i]);
            s = (i-5) < 0   ? 0   : (i-5);
            e = (i+5) > len ? len : (i+5);
            printf ("[");
            for (i = s; i < e; i++)
                printf ("%c", buf1[i]);
            printf ("]\n[");
            for (i = s; i < e; i++)
                printf ("%c", buf2[i]);
            printf ("]\n");
            return 1;
        }
    }
    return 0;
}


INT stripbuf (CHAR *buf, INT len)
{
    CHAR *ptr;
    INT count;

    for (ptr = buf, count = 0; len-- > 0; buf++) {
        if (*buf == '\r' && *(buf+1) == '\n')
            continue;
        *ptr++ = *buf;
        count++;
    } 
    return count;
}


INT main (INT argc, CHAR *argv[])
{
#define NS 100

    INT step = 0, s, ns;
    CHAR segbuf[2*SEGSIZE], cmpbuf[2*SEGSIZE];
    INT lens[NS], err = 0, rtc; 
    
    MsgLevel  = 5;
    MaxSyslog = -1;
    startLog();

    printf ("Step %d: Writing a binary file of %d full segments\n", 
             ++step, ns = NS);
    mkTestfile();
    for (s = 0; s < ns; s++) {
        lens[s] = SEGSIZE;
        filbuf(segbuf, lens[s], s);
        rtc = putXferBuffer (testfd, segbuf, lens[s], RM_OCTET, (s == (ns-1)) ? 1 : 0, NULL);
        if (rtc != SEGSIZE) {
            printf ("Error writing segment %d\n", s);
            err = 1;
            break;
        }
    }
    rewind (testfd);
    if (!err) {
        for (s = 0; s < ns; s++) {
            filbuf(cmpbuf, lens[s], s);
            if (fread(segbuf, 1, lens[s], testfd) != lens[s]) {
                printf ("Error reading segment %d\n", s);
                err = 1;
                break;
            }
            if ((err = docompare (s, segbuf, cmpbuf, lens[s])))
                break;
        }
    }
    removeTestfile();

    printf ("Step %d: Writing a binary file of 1 half and %d full segments\n",
             ++step, (ns = NS) -1 );
    mkTestfile();
    for (s = 0; s < ns; s++) {
        lens[s] = (s) ? SEGSIZE : SEGSIZE/2;
        filbuf(segbuf, lens[s], s);
        rtc = putXferBuffer (testfd, segbuf, lens[s], RM_OCTET, (s == (ns-1)) ? 1 : 0, NULL);
        if (rtc != lens[s]) {
            printf ("Error writing segment %d\n", s);
            err = 1;
            break;
        }
    }
    rewind (testfd);
    if (!err) {
        for (s = 0; s < ns; s++) {
            filbuf(cmpbuf, lens[s], s);
            if (fread(segbuf, 1, lens[s], testfd) != lens[s]) {
                printf ("Error reading segment %d\n", s);
                err = 1;
                break;
            }
            if ((err = docompare (s, segbuf, cmpbuf, lens[s])))
                break;
        }
    }
    removeTestfile();

    printf ("Step %d: Reading a binary file of %d full segments\n", 
             ++step, ns = NS);
    mkTestfile();
    for (s = 0; s < ns; s++) {
        lens[s] = SEGSIZE;
        filbuf(segbuf, lens[s], s);
        if (fwrite(segbuf, 1, lens[s], testfd) != lens[s]) {
            printf ("Error writing segment %d\n", s);
            err = 1;
            break;
        }
    }
    rewind (testfd);
    if (!err) {
        for (s = 0; s < ns; s++) {
            filbuf(cmpbuf, lens[s], s);
            rtc = getXferBuffer (testfd, segbuf, lens[s], RM_OCTET, NULL);
            if (rtc != lens[s]) {
                printf ("Error reading segment %d\n", s);
                err = 1;
                break;
            }
            if ((err = docompare (s, segbuf, cmpbuf, lens[s])))
                break;
        }
    }
    if(getXferBuffer (testfd, segbuf, lens[s], RM_OCTET, NULL))
        printf ("No EOF -- dangling test file data");
    removeTestfile();

    printf ("Step %d: Reading a binary file of %d + 1/2 segments\n", 
             ++step, (ns = NS) -1);
    mkTestfile();
    for (s = 0; s < ns; s++) {
        lens[s] = (ns == (NS-1)) ? SEGSIZE/2 : SEGSIZE;
        filbuf(segbuf, lens[s], s);
        if (fwrite(segbuf, 1, lens[s], testfd) != lens[s]) {
            printf ("Error writing segment %d\n", s);
            err = 1;
            break;
        }
    }
    rewind (testfd);
    if (!err) {
        for (s = 0; s < ns; s++) {
            filbuf(cmpbuf, lens[s], s);
            rtc = getXferBuffer (testfd, segbuf, SEGSIZE, RM_OCTET, NULL);
            if (rtc != lens[s]) {
                printf ("Error reading segment %d\n", s);
                err = 1;
                break;
            }
            if ((err = docompare (s, segbuf, cmpbuf, lens[s])))
                break;
        }
    }
    if(getXferBuffer (testfd, segbuf, lens[s], RM_OCTET, NULL))
        printf ("No EOF -- dangling test file data");
    removeTestfile();

    printf ("Step %d: Writing a file with netascii conversion of %d full segments\n", 
             ++step, ns = NS);
    mkTestfile();
    for (s = 0; s < ns; s++) {
        lens[s] = SEGSIZE;
        filbuf(segbuf, lens[s], s);
        rtc = putXferBuffer (testfd, segbuf, lens[s], RM_NETASCII, (s == (ns-1)) ? 1 : 0, NULL);
        if (rtc != SEGSIZE) {
            printf ("Error writing segment %d\n", s);
            err = 1;
            break;
        }
    }
    rewind (testfd);
    if (!err) {
        for (s = 0; s < ns; s++) {
            filbuf(cmpbuf, lens[s], s);
            lens[s] = stripbuf(cmpbuf, lens[s]);
            if (fread(segbuf, 1, lens[s], testfd) != lens[s]) {
                printf ("Error reading segment %d\n", s);
                err = 1;
                break;
            }
            if ((err = docompare (s, segbuf, cmpbuf, lens[s])))
                break;
        }
    }
    removeTestfile();

    printf ("Step %d: Reading a file with netascii conversion of %d full segments\n", 
             ++step, ns = NS);
    mkTestfile();
    for (s = 0; s < ns; s++) {
        lens[s] = SEGSIZE;
        filbuf(segbuf, lens[s], s);
        lens[s] = stripbuf(segbuf, lens[s]);
        if (fwrite(segbuf, 1, lens[s], testfd) != lens[s]) {
            printf ("Error writing segment %d\n", s);
            err = 1;
            break;
        }
    }
    rewind (testfd);
    if (!err) {
        for (s = 0; s < ns; s++) {
            lens[s] = SEGSIZE;
            filbuf(cmpbuf, lens[s], s);
            rtc = getXferBuffer (testfd, segbuf, lens[s], RM_NETASCII, NULL);
            if (rtc != lens[s]) {
                printf ("Error reading segment %d\n", s);
                err = 1;
                break;
            }
            if ((err = docompare (s, segbuf, cmpbuf, lens[s])))
                break;
        }
    }
    if(getXferBuffer (testfd, segbuf, lens[s], RM_OCTET, NULL))
        printf ("No EOF -- dangling test file data");
    removeTestfile();


    return 0;
}

#endif /* TEST */
